from datetime import date
from typing import Optional

from tastyworks.models.option_chain import get_option_chain
from tastyworks.models.underlying import Underlying

from ..utils import RenewableTastyAPISession, get_tasty_monthly, is_monthly


async def choose_expiration(sesh: RenewableTastyAPISession, undl: Underlying,
                            all_expirations: Optional[bool] = False) -> date:
    chain = await get_option_chain(sesh, undl)
    exps = chain.get_all_expirations()
    if not all_expirations:
        exps = [exp for exp in exps if is_monthly(exp)]
    default = get_tasty_monthly()
    for i in range(len(exps)):
        if exps[i] == default:
            print(f'{i + 1}) {exps[i]} (default)')
        else:
            print(f'{i + 1}) {exps[i]}')
    choice = 0
    while choice not in range(1, len(exps) + 1):
        try:
            raw = input('Please choose an expiration: ')
            choice = int(raw)
        except ValueError:
            if not raw:
                return default

    return exps[choice - 1]



"""Funcs to calculate size of pair"""


def size(price_left=1, price_right=1, vol_left=1, vol_right=1, unit_size_left=1,
         multiplier_left=1, multiplier_right=1):
    """Calculate volatility weighted size for each leg of pairs trade

    Calculate the number of units for each leg of a pairs trade by
    passing the price of both legs and optionally the share size of one
    leg and multipliers in the case of futures. Returns a dict with the
    given arguments and the number of units for each leg.

    It is often desirable to equate the level of risk for each leg of a
    pairs trade. This risk depends on both the equity value and the
    volatility of each leg.

    For example, suppose a pairs trade consists of two legs. The first
    leg is long stock ABC, which trades for $100 per share and has a
    30-day historical standard deviation of 2%. The second leg is a
    short position in stock XYZ, which trades for $25 per share and has
    a standard deviation of 7%. How many shares of each stock should be
    traded to create a neutral pairs trade?

    Clearly, if 100 shares of each leg were traded, the risk in one leg
    would be much greater than the other. So, we want to equate the
    equity of each leg:

        abc_shares * abc_price = xyz_shares * xyz_price
        left_shares * left_price = right_shares * right_price

    ...where we have arbitrarily defined ABC as the "left" side of the
    trade.

    However, the above equation fails to adequately equate the risk in
    each leg; since the volatility of XYZ is 7% and the volatility of
    ABC is 2%, sizing the trade based on the above equation would put
    more risk in the XYZ leg than the ABC leg.

    So, to equate the risk in each leg, we add a measure of volatility
    to the prior equation:

        left_shares * left_price * left_vol = right_shares * right_price * right_vol

    Compared to the prior equation, this equation reduces the number of
    shares traded in XYZ, since the risk is larger in each unit of
    equity for XYZ, because XYZ has higher volatility.

    Now, suppose that we're comfortable trading 100 shares of ABC and we
    want to now calculate the quantity of XYZ shares to trade. We can do
    a simple algebriac manipulation to calculate the number of shares:

        right_shares = (left_shares * left_price * left_vol) / (right_price * right_vol)
                     = (100 * 100 * 2) / (25 * 7)
                     = 114.28

    ..and this can be plugged back into the original equation as a
    check.

    In the case of a futures contract, the *notional value* must be
    calculated in order to volatility-weight each leg:

        left_units * left_multiplier * left_price * left_vol =
            right_units * right_multiplier * right_price * right_vol

    ...where the multiplier is a number used to convert the futures
    price into the notional value of the futures contract. In the case
    where both legs are a stock, the multipliers are simply 1. To
    calculate the number of units for the right leg:

        left_units =
            (right_units * right_multiplier * right_price * right_vol) /
            (left_multiplier * left_price * left_vol)

    Parameters
    ----------
    price_left, price_right : float
        The entry prices for each leg of the pair, left and right.

    vol_left, vol_right : float
        A measure of volatility for each leg of the pair, left and
        right.

    unit_size_left : float
        The number of units (e.g. shares) that compose the left leg of
        the pair. Optional, default is 1 unit.

    multiplier_left, multiplier_right : float
        Multiplier for each leg of the pair, left and right. The
        multiplier is a number that translates the asset price into the
        notional (equity) value of one unit of the asset. In the case of
        a stock, the mulitplier is simply 1. The multiplier for futures
        contracts depends on the specifications of the contract.
        Optional, default is 1.

    Returns
    -------
    pair_specs : dict
        Dict with keys multiplier_left, multiplier_right, price_left,
        price_right, unit_size_left, unit_size_right, vol_left,
        and vol_right.

    """
    unit_size_right = \
        (vol_left * price_left * multiplier_left * unit_size_left) / \
        (vol_right * price_right * multiplier_right)

    pair_specs = dict(vol_left=vol_left, price_left=price_left,
                      multiplier_left=multiplier_left,
                      multiplier_right=multiplier_right,
                      unit_size_left=unit_size_left, vol_right=vol_right,
                      price_right=price_right, unit_size_right=unit_size_right)

    return pair_specs



from typing import Optional

import asyncclick as click
import petl

from ..utils import RenewableTastyAPISession, get_account
from .plot import Portfolio


@click.command(help='Chart your net liquidity or realized profit/loss over time.')
@click.option('-n', '--netliq', is_flag=True,
              help='Display net liquidity over time instead of realized profit/loss.')
@click.option('-p', '--percentage', is_flag=True,
              help='Whether to display percentages instead of absolute values in the chart.')
@click.option('-d', '--duration', default='ytd',
              help='Possible values: {all,10y,5y,1y,ytd,6m,3m,1m,5d}')
async def plot(netliq: Optional[bool] = False, percentage: Optional[bool] = False, duration: Optional[str] = None):
    sesh = await RenewableTastyAPISession.create()
    # choose an account
    acc = await get_account(sesh)
    history = await acc.get_history(sesh)

    table = petl.fromdicts(history).cut(
        'executed-at',
        'transaction-type',
        'action',
        'symbol',
        'value',
        'value-effect',
        'quantity',
        'commission',
        'clearing-fees',
        'proprietary-index-option-fees',
        'regulatory-fees'
    ).addfield('is-closing', lambda row: 'Close' in row['action'] if row['action'] else False) \
     .sort(['executed-at', 'is-closing'])

    # create a portfolio with the given history
    pf = Portfolio(petl.data(table), net_liq=netliq)

    # get initial net liq if we're using percentage
    nl = None
    if percentage:
        pf_tmp = Portfolio(petl.data(table), net_liq=True)
        nl = pf_tmp._get_starting_net_liq(duration)

    # get the P/L or net liq and save the graph
    val = pf.plot(duration, starting_net_liq=nl)

    # print current positions
    if nl is None:
        print(('Current net liquidity' if netliq else 'Realized P/L') + f': ${val:.2f}')
    else:
        print(('Change in net liquidity' if netliq else 'Realized P/L') + f': {val:.2f}%')
    print('Current positions:')
    for p in pf.positions.values():
        print(p)



import os
import subprocess
import sys
from datetime import datetime
from decimal import Decimal

import matplotlib.pyplot as plt
from dateutil.relativedelta import relativedelta
from matplotlib.dates import (DateFormatter, DayLocator, MonthLocator,
                              YearLocator)

from ..utils import ZERO, TastyworksCLIError

_month_fmt = DateFormatter('%b')
_year_fmt = DateFormatter('%Y')
_day_fmt = DateFormatter('%d')

_DURATIONS = ['all', '10y', '5y', '1y', 'ytd', '6m', '3m', '1m', '5d']


class Trade():
    '''
    Object to hold a specific trade and its notable properties.
    Contains the important information of one row of the table.
    '''
    def __init__(self, trade):
        self.date = trade[0].split('T')[0]  # we just want the date
        self.type = trade[1]
        self.action = trade[2]
        self.symbol = trade[3]
        self.value = Decimal(trade[4].replace(',', '')) * (-1 if trade[5] == 'Debit' else 1)
        self.quantity = Decimal(trade[6].replace(',', '')) if trade[6] else ZERO
        commission = -Decimal(trade[7]) if trade[7] else ZERO
        clearing_fees = -Decimal(trade[8]) if trade[8] else ZERO
        pio_fees = -Decimal(trade[9]) if trade[9] else ZERO
        regulatory_fees = -Decimal(trade[10]) if trade[10] else ZERO
        self.fees = commission + clearing_fees + pio_fees + regulatory_fees

    def __str__(self):
        return f'{self.date}: {self.symbol} x{self.quantity} at ${self.value}'


class Portfolio():
    '''
    Starting with the given table, runs forwards through
    time, tracking either realized P/L or net liquidity,
    and saving those results. Contains methods to process
    a petl table, handle date ranges, and draw a plot.
    '''
    def __init__(self, json, net_liq=False):
        # contains the P/L or net liquidity at a certain date
        self.values = []
        # contains the dates at which closing trades were placed
        self.dates = []
        # a dictionary of symbol -> Trade, used to close positions
        self.positions = {}
        self.last_value = ZERO
        # the given petl table
        self.json = json
        # whether or not to use net liq instead of realized P/L
        self.net_liq = net_liq

        self._calculate()

    def _process_dates(self, duration):
        if duration not in _DURATIONS:
            raise TastyworksCLIError('Not a valid duration!\t{all,10y,5y,1y,ytd,6m,3m,1m,5d}')

        # convert dates to datetimes
        self.dates = [datetime.strptime(date, '%Y-%m-%d') for date in self.dates]

        # when multiple transactions occur in a day, use only the last one
        final_dates = []
        final_price = []
        last_date = None
        for i in range(len(self.dates)):
            if last_date is None:
                final_dates.append(self.dates[i])
                final_price.append(self.values[i])
            elif self.dates[i] != last_date:
                final_dates.append(last_date)
                final_price.append(self.values[i - 1])
            last_date = self.dates[i]
        # method above omits the very last trade
        final_dates.append(last_date)
        final_price.append(self.values[-1])

        # update these lists to the cleaned versions
        self.dates = final_dates
        self.values = final_price

        if duration == 'all':
            start_date = self.dates[0]
        elif duration == '10y':
            start_date = datetime.now() - relativedelta(years=10)
        elif duration == '5y':
            start_date = datetime.now() - relativedelta(years=5)
        elif duration == '1y':
            start_date = datetime.now() - relativedelta(years=1)
        elif duration == 'ytd':
            # hack to get first day of current year in a datetime
            start_date = datetime.combine(
                datetime.now().date().replace(month=1, day=1),
                datetime.min.time()
            )
        elif duration == '6m':
            start_date = datetime.now() - relativedelta(months=6)
        elif duration == '3m':
            start_date = datetime.now() - relativedelta(months=3)
        elif duration == '1m':
            start_date = datetime.now() - relativedelta(months=1)
        else:
            start_date = datetime.now() - relativedelta(days=7)

        # if given range is longer than our portfolio history
        if start_date < self.dates[0]:
            start_date = self.dates[0]

        # get index of first date in our range
        start = -1
        for i in range(len(self.dates)):
            if self.dates[i] >= start_date:
                start = i
                break

        if self.dates[start] == self.dates[-1]:
            raise TastyworksCLIError('Not enough closing trades present! There must be at least two closing trades present in the given timeframe.')

        # change formatting based on scale
        delta = datetime.now() - self.dates[0]
        if delta.days <= 31:
            fun = _d_fmt
            loc = DayLocator()
        elif delta.days <= 365:
            fun = _m_fmt_long
            loc = MonthLocator()
        elif delta.days <= 1000:
            fun = _m_fmt_short
            loc = MonthLocator()
        else:
            fun = _y_fmt
            loc = YearLocator()

        return start, fun, loc

    def _get_starting_net_liq(self, duration):
        '''
        Calculates the net liquidity at the beginning of
        the given time period for use with the percentage
        argument. Modifies the state, so call this on a
        throwaway Portfolio instance.
        '''
        start, _, _ = self._process_dates(duration.lower())

        return self.values[start]

    def plot(self, duration, starting_net_liq=None, gen_img=True):
        start, fun, loc = self._process_dates(duration.lower())

        # graph percentages
        if starting_net_liq is not None:
            initial_value = self.values[start]
            for i in range(start, len(self.values)):
                self.values[i] = (self.values[i] - initial_value) / starting_net_liq * Decimal(100)
        # shift graph vertically so it starts at zero if doing P/L
        elif not self.net_liq:
            initial_value = self.values[start]
            for i in range(start, len(self.values)):
                self.values[i] -= initial_value

        if gen_img:
            fig, ax = plt.subplots()

            # color based on net liq or profitability
            if self.net_liq:
                color = 'steelblue'
            elif self.values[-1] < 0:
                color = 'crimson'  # :(
            else:
                color = 'mediumseagreen'

            plt.plot(self.dates[start:], self.values[start:], color=color)
            plt.title('Net Liquidity' if self.net_liq else 'Realized P/L')
            ax.xaxis.set_major_locator(loc)
            ax.xaxis.set_major_formatter(fun)

            # save plot to current directory
            fp = 'netliq.png' if self.net_liq else 'pandl.png'
            fig.savefig(fp)

            # open plot in default image viewer
            if sys.platform == 'win32':
                os.startfile(fp)
            else:
                opener = 'open' if sys.platform == 'darwin' else 'xdg-open'
                subprocess.call([opener, fp])

        # return either the final net liq or the change in P/L
        return (self.values[-1] if self.net_liq else self.values[-1] - self.values[start])

    def _calculate(self):
        for trade in self.json:
            t = Trade(trade)
            # we could allow the user to disable this adjustment
            self.last_value += t.fees

            if t.type == 'Trade':
                # futures are handled differently. instead of opening
                # or closing trades, you just have a number of long
                # or short contracts. we multiply short contract quantities
                # by -1 and whenever a symbol's number of contracts hits
                # 0, it's treated as realized.
                if t.symbol[0] == '/':
                    # symbol already in positions dict
                    if t.symbol in self.positions:
                        self.positions[t.symbol].quantity += t.quantity * (-1 if 'sell' in t.action.lower() else 1)
                        self.positions[t.symbol].value += t.value
                    else:
                        self.positions[t.symbol] = t
                        self.positions[t.symbol].quantity *= (-1 if 'sell' in t.action.lower() else 1)

                    # realize gain/loss here
                    if self.positions[t.symbol].quantity == 0:
                        self.dates.append(t.date)
                        self.last_value += self.positions[t.symbol].value
                        self.values.append(self.last_value)
                        del self.positions[t.symbol]
                # non-futures opening trades
                elif 'open' in t.action.lower():
                    # symbol already in positions dict
                    if t.symbol in self.positions:
                        self.positions[t.symbol].quantity += t.quantity
                        self.positions[t.symbol].value += t.value
                    else:
                        self.positions[t.symbol] = t
                # non-futures closing trades
                else:
                    if t.symbol not in self.positions:
                        raise TastyworksCLIError(f'Closing trade present but opening trade missing for trade:\n{t}')
                    else:
                        self.positions[t.symbol].quantity -= t.quantity
                        self.positions[t.symbol].value += t.value
                        # realize if position is *completely* closed out
                        if self.positions[t.symbol].quantity == 0:
                            self.dates.append(t.date)
                            # this brings up an interesting question:
                            # is this a good way to determine realized/unrealized?
                            # example: you buy 50 shares at $10 and 50 shares at $11.
                            # then you sell 50 shares at $12. should that count as:
                            # unrealized until whole position is closed? (current)
                            # realized at the average cost basis?
                            # or realized at a specific basis?
                            self.last_value += self.positions[t.symbol].value  # the line in question
                            self.values.append(self.last_value)
                            del self.positions[t.symbol]

            # TW uses this for several things. the ones we care
            # about are expiration, assignment, exercise, and awards.
            elif t.type == 'Receive Deliver':
                # exercise/awards
                if t.action == 'Buy to Open' or t.action == 'Sell to Open':
                    if t.symbol in self.positions:
                        self.positions[t.symbol].quantity += t.quantity
                        self.positions[t.symbol].value += t.value
                    else:
                        self.positions[t.symbol] = t

                # assignment/expiration/symbol change
                elif t.symbol in self.positions:
                    self.positions[t.symbol].quantity -= t.quantity
                    self.positions[t.symbol].value += t.value
                    # realize if position is closed
                    if self.positions[t.symbol].quantity == 0:
                        self.dates.append(t.date)
                        self.last_value += self.positions[t.symbol].value
                        self.values.append(self.last_value)
                        del self.positions[t.symbol]

            # apply mark-to-market even when not tracking net liq;
            # otherwise, this is only done if we're tracking net liq.
            elif t.type == 'Money Movement' and (self.net_liq or (t.symbol[0] == '/' if t.symbol else False)):
                self.dates.append(t.date)
                self.last_value += t.value
                self.values.append(self.last_value)


def _m_fmt_long(x, pos=None):
    if _month_fmt(x) == 'Jan':
        return _year_fmt(x)
    return _month_fmt(x)


def _m_fmt_short(x, pos=None):
    if _month_fmt(x) == 'Jan':
        return _year_fmt(x)[-2:]
    return _month_fmt(x)[0]


def _y_fmt(x, pos=None):
    return _year_fmt(x)


def _d_fmt(x, pos=None):
    if _day_fmt(x) == '01':
        return _month_fmt(x)
    return _day_fmt(x)



from decimal import Decimal
from typing import Optional

import asyncclick as click
from rich.console import Console
from rich.table import Table
from tastyworks.models.option import Option, OptionType
from tastyworks.models.option_chain import get_option_chain
from tastyworks.models.order import (Order, OrderDetails, OrderPriceEffect,
                                     OrderType)
from tastyworks.models.underlying import Underlying, UnderlyingType
from tastyworks.streamer import DataStreamer

from ..utils import (RenewableTastyAPISession, TastyworksCLIError, get_account,
                     get_confirmation)
from .option import choose_expiration


@click.group(chain=True, help='Buy, sell, and analyze options.')
async def option():
    pass


@option.command(help='Buy or sell calls with the given parameters.')
@click.option('-s', '--strike', type=int, help='The chosen strike for the option.')
@click.option('-d', '--delta', type=int, help='The chosen delta for the option.')
@click.option('--all-expirations', is_flag=True, help='Show all expirations, not just monthlies.')
@click.argument('underlying', type=str)
@click.argument('quantity', type=int)
async def call(underlying: str, quantity: int, strike: Optional[int] = None,
               all_expirations: Optional[bool] = False, delta: Optional[int] = None):
    if strike is not None and delta is not None:
        raise TastyworksCLIError('Must specify either delta or strike, but not both.')
    elif not strike and not delta:
        raise TastyworksCLIError('Please specify either delta or strike for the option.')
    elif abs(delta) > 100:
        raise TastyworksCLIError('Delta value is too high, -100 <= delta <= 100')

    sesh = await RenewableTastyAPISession.create()
    undl = Underlying(underlying)
    expiration = await choose_expiration(sesh, undl, all_expirations)
    streamer = await DataStreamer.create(sesh)

    if not strike:
        chain = await get_option_chain(sesh, undl, expiration)
        dxfeeds = [option.symbol_dxf for option in chain.options if option.option_type == OptionType.CALL]
        greeks = await streamer.stream('Greeks', dxfeeds)

        lowest = abs(greeks[0]['delta'] * 100.0 - delta)
        index = 0
        for i in range(1, len(greeks)):
            diff = abs(greeks[i]['delta'] * 100.0 - delta)
            if diff < lowest:
                index = i
                lowest = diff
        for option in chain.options:
            if option.symbol_dxf == greeks[index]['eventSymbol']:
                strike = option.strike
                break

    option = Option(
        ticker=underlying,
        quantity=abs(quantity),
        expiry=expiration,
        strike=strike,
        option_type=OptionType.CALL,
        underlying_type=UnderlyingType.EQUITY
    )
    quote = await streamer.stream('Quote', [option.symbol_dxf])
    bid = quote[0]['bidPrice']
    ask = quote[0]['askPrice']
    mid = (bid + ask) / 2

    await streamer.close()

    console = Console()
    table = Table(show_header=True, header_style='bold', title_style='bold',
                  title=f'Quote for {underlying} {strike}C {expiration}')
    table.add_column('Bid', style='green', width=8, justify='center')
    table.add_column('Mid', width=8, justify='center')
    table.add_column('Ask', style='red', width=8, justify='center')
    table.add_row(f'{bid:.2f}', f'{mid:.2f}', f'{ask:.2f}')
    console.print(table)

    price = input('Please enter a limit price for the entire order (default mid): ')
    if not price:
        price = round(mid * abs(quantity), 2)
    price = Decimal(price)

    details = OrderDetails(
        type=OrderType.LIMIT,
        price=price,
        price_effect=OrderPriceEffect.CREDIT if quantity < 0 else OrderPriceEffect.DEBIT
    )
    order = Order(details)
    order.add_leg(option)

    acct = await get_account(sesh)
    details = await acct.get_balance(sesh)
    nl = Decimal(details['net-liquidating-value'])

    data = await acct.execute_order(order, sesh, dry_run=True)
    bp = Decimal(data['buying-power-effect']['change-in-buying-power'])
    percent = bp / nl * Decimal(100)
    # bp_effect = data['buying-power-effect']['change-in-buying-power-effect']
    fees = Decimal(data['fee-calculation']['total-fees'])

    table = Table(show_header=True, header_style='bold', title_style='bold', title='Order Review')
    table.add_column('Quantity', width=8, justify='center')
    table.add_column('Symbol', width=8, justify='center')
    table.add_column('Strike', width=8, justify='center')
    table.add_column('Type', width=8, justify='center')
    table.add_column('Expiration', width=10, justify='center')
    table.add_column('Price', width=8, justify='center')
    table.add_column('BP', width=8, justify='center')
    table.add_column('% of NL', width=8, justify='center')
    table.add_column('Fees', width=8, justify='center')
    table.add_row(f'{quantity}', underlying, f'{strike:.2f}', 'CALL', f'{expiration}', f'${price:.2f}',
                  f'${bp:.2f}', f'{percent:.2f}%', f'${fees}')
    console.print(table)

    if data['warnings']:
        console.print('[bold orange]Warnings:[/bold orange]')
        for warning in data['warnings']:
            console.print(f'[i gray]{warning}[/i gray]')
    if get_confirmation('Send order? Y/n '):
        await acct.execute_order(order, sesh, dry_run=False)


@option.command(help='Buy or sell puts with the given parameters.')
@click.option('-s', '--strike', type=int, help='The chosen strike for the option.')
@click.option('-d', '--delta', type=int, help='The chosen delta for the option.')
@click.option('--all-expirations', is_flag=True, help='Show all expirations, not just monthlies.')
@click.argument('underlying', type=str)
@click.argument('quantity', type=int)
async def put(underlying: str, quantity: int, strike: Optional[int] = None,
              all_expirations: Optional[bool] = False, delta: Optional[int] = None):
    if strike is not None and delta is not None:
        raise TastyworksCLIError('Must specify either delta or strike, but not both.')
    elif not strike and not delta:
        raise TastyworksCLIError('Please specify either delta or strike for the option.')
    elif abs(delta) > 100:
        raise TastyworksCLIError('Delta value is too high, -100 <= delta <= 100')

    sesh = await RenewableTastyAPISession.create()
    undl = Underlying(underlying)
    expiration = await choose_expiration(sesh, undl, all_expirations)
    streamer = await DataStreamer.create(sesh)

    if not strike:
        chain = await get_option_chain(sesh, undl, expiration)
        dxfeeds = [option.symbol_dxf for option in chain.options if option.option_type == OptionType.PUT]
        greeks = await streamer.stream('Greeks', dxfeeds)

        lowest = abs(greeks[0]['delta'] * 100.0 + delta)
        index = 0
        for i in range(1, len(greeks)):
            diff = abs(greeks[i]['delta'] * 100.0 + delta)
            if diff < lowest:
                index = i
                lowest = diff
        for option in chain.options:
            if option.symbol_dxf == greeks[index]['eventSymbol']:
                strike = option.strike
                break

    option = Option(
        ticker=underlying,
        quantity=abs(quantity),
        expiry=expiration,
        strike=strike,
        option_type=OptionType.PUT,
        underlying_type=UnderlyingType.EQUITY
    )
    quote = await streamer.stream('Quote', [option.symbol_dxf])
    bid = quote[0]['bidPrice']
    ask = quote[0]['askPrice']
    mid = (bid + ask) / 2

    await streamer.close()

    console = Console()
    table = Table(show_header=True, header_style='bold', title_style='bold',
                  title=f'Quote for {underlying} {strike}P {expiration}')
    table.add_column('Bid', style='green', width=8, justify='center')
    table.add_column('Mid', width=8, justify='center')
    table.add_column('Ask', style='red', width=8, justify='center')
    table.add_row(f'{bid:.2f}', f'{mid:.2f}', f'{ask:.2f}')
    console.print(table)

    price = input('Please enter a limit price for the entire order (default mid): ')
    if not price:
        price = round(mid * abs(quantity), 2)
    price = Decimal(price)

    details = OrderDetails(
        type=OrderType.LIMIT,
        price=price,
        price_effect=OrderPriceEffect.CREDIT if quantity < 0 else OrderPriceEffect.DEBIT
    )
    order = Order(details)
    order.add_leg(option)

    acct = await get_account(sesh)
    details = await acct.get_balance(sesh)
    nl = Decimal(details['net-liquidating-value'])

    data = await acct.execute_order(order, sesh, dry_run=True)
    bp = Decimal(data['buying-power-effect']['change-in-buying-power'])
    percent = bp / nl * Decimal(100)
    # bp_effect = data['buying-power-effect']['change-in-buying-power-effect']
    fees = Decimal(data['fee-calculation']['total-fees'])

    table = Table(show_header=True, header_style='bold', title_style='bold', title='Order Review')
    table.add_column('Quantity', width=8, justify='center')
    table.add_column('Symbol', width=8, justify='center')
    table.add_column('Strike', width=8, justify='center')
    table.add_column('Type', width=8, justify='center')
    table.add_column('Expiration', width=10, justify='center')
    table.add_column('Price', width=8, justify='center')
    table.add_column('BP', width=8, justify='center')
    table.add_column('% of NL', width=8, justify='center')
    table.add_column('Fees', width=8, justify='center')
    table.add_row(f'{quantity}', underlying, f'{strike:.2f}', 'PUT', f'{expiration}', f'${price:.2f}',
                  f'${bp:.2f}', f'{percent:.2f}%', f'${fees}')
    console.print(table)

    if data['warnings']:
        console.print('[bold orange]Warnings:[/bold orange]')
        for warning in data['warnings']:
            console.print(f'[i gray]{warning}[/i gray]')
    if get_confirmation('Send order? Y/n '):
        await acct.execute_order(order, sesh, dry_run=False)


@option.command(help='Fetch and display an options chain.')
@click.option('-s', '--strikes', type=int, default=8,
              help='The number of strikes to fetch above and below the spot price.')
@click.argument('underlying', type=str)
async def chain(underlying: str, strikes: Optional[int] = None):
    sesh = await RenewableTastyAPISession.create()
    undl = Underlying(underlying)

    strike_price = 0
    streamer = await DataStreamer.create(sesh)
    quote = await streamer.stream('Quote', [underlying])
    bid = quote[0]['bidPrice']
    ask = quote[0]['askPrice']
    strike_price = (bid + ask) / 2

    expiration = await choose_expiration(sesh, undl)

    console = Console()
    table = Table(show_header=True, header_style='bold', title_style='bold',
                  title=f'Options chain for {underlying} on {expiration}')
    table.add_column('Delta', width=8, justify='center')
    table.add_column('Bid', style='red', width=8, justify='center')
    table.add_column('Ask', style='red', width=8, justify='center')
    table.add_column('Strike', width=8, justify='center')
    table.add_column('Bid', style='green', width=8, justify='center')
    table.add_column('Ask', style='green', width=8, justify='center')
    table.add_column('Delta', width=8, justify='center')

    chain = await get_option_chain(sesh, undl, expiration)
    puts = []
    calls = []
    for option in chain.options:
        if option.option_type == OptionType.CALL:
            calls.append(option)
        else:
            puts.append(option)

    puts_to_fetch = []
    calls_to_fetch = []
    if strikes * 2 < len(calls):
        mid_index = 0
        while calls[mid_index].strike < strike_price:
            mid_index += 1
        # handle weird edge scenarios here
        for call in calls[mid_index - strikes:mid_index + strikes]:
            calls_to_fetch.append(call)
        for put in puts[mid_index - strikes:mid_index + strikes]:
            puts_to_fetch.append(put)
    else:
        for call in calls:
            calls_to_fetch.append(call)
        for put in puts:
            puts_to_fetch.append(put)

    dxfeeds = [call.symbol_dxf for call in calls_to_fetch] + [put.symbol_dxf for put in puts_to_fetch]
    quotes = await streamer.stream('Quote', dxfeeds)
    greeks = await streamer.stream('Greeks', dxfeeds)

    for i in range(len(calls_to_fetch)):
        call_dxf = calls_to_fetch[i].symbol_dxf
        put_dxf = puts_to_fetch[i].symbol_dxf
        strike = calls_to_fetch[i].strike
        put_bid = 0
        put_ask = 0
        put_delta = 0
        call_bid = 0
        call_ask = 0
        call_delta = 0
        for item in quotes:
            if item['eventSymbol'] == put_dxf:
                put_bid = item['bidPrice']
                put_ask = item['askPrice']
                break
        for item in greeks:
            if item['eventSymbol'] == put_dxf:
                put_delta = int(item['delta'] * 100)
                break
        for item in quotes:
            if item['eventSymbol'] == call_dxf:
                call_bid = item['bidPrice']
                call_ask = item['askPrice']
                break
        for item in greeks:
            if item['eventSymbol'] == call_dxf:
                call_delta = int(item['delta'] * 100)
                break

        table.add_row(
            f'{put_delta:g}',
            f'{put_bid:.2f}',
            f'{put_ask:.2f}',
            f'{strike:.2f}',
            f'{call_bid:.2f}',
            f'{call_ask:.2f}',
            f'{call_delta:g}'
        )
        if i == strikes - 1:
            table.add_row('=======', 'ITM v', '=======', '=======', '=======', 'ITM ^', '=======', style='white')

    await streamer.close()
    console.print(table)



import asyncio
import logging
import sys

import asyncclick as click

from .option.commands import option
from .plot.commands import plot
from .utils import VERSION

LOGGER = logging.getLogger(__name__)


@click.group()
@click.version_option(VERSION)
async def app():
    pass


def main():
    if sys.version_info[0] == 3 and sys.version_info[1] >= 8 and sys.platform.startswith('win'):
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        LOGGER.debug('Using Windows-specific event loop policy')

    app.add_command(option)
    app.add_command(plot)

    app(_anyio_backend='asyncio')



"""Convenience functions, class for optons price modeling with QuantLib"""
import pandas as pd
import QuantLib as ql


def summarize_params(price_underlying, strike, interest, volatility, date_exp,
                     date_evl):
    """Display values that are inputs to option price"""
    msg = (f"* {price_underlying.value()} - underlying price\n"
           f"* {strike} - strike\n"
           f"* {interest} - interest\n"
           f"* {volatility.value()} - volatility\n"
           f"* {date_exp} - exp date\n"
           f"* {date_evl} - eval date\n")

    print(msg)


def build_volatility_curve(volatility):
    """Create volatility curve"""
    volatility = ql.SimpleQuote(volatility)
    volatility_curve = ql.BlackConstantVol(0, ql.TARGET(),
                                           ql.QuoteHandle(volatility),
                                           ql.Actual360())
    return volatility, volatility_curve


def build_interest_curve(rate_risk_free):
    """Create interest curve"""
    rate_risk_free = ql.SimpleQuote(rate_risk_free)
    rate_risk_free_curve = ql.FlatForward(0, ql.TARGET(),
                                          ql.QuoteHandle(rate_risk_free),
                                          ql.Actual360())

    return rate_risk_free_curve


def build_dates(date_expiration=None, date_evaluation=None, dte=None):
    """Create date objects for model input"""
    # setup dates
    if not date_expiration and not date_evaluation:
        date_evaluation = pd.Timestamp.utcnow().tz_convert('US/Central')
        date_expiration = date_evaluation + dte
    elif not dte and ((not date_expiration) or (not date_evaluation)):
        raise Exception('Must have either both exp and eval dates or days to '
                        'expiration. ')
    else:
        if date_evaluation == 'today':
            date_evaluation = pd.Timestamp.utcnow().tz_convert('US/Central')
        date_expiration = pd.to_datetime(date_expiration)
        date_evaluation = pd.to_datetime(date_evaluation)

    dte = (date_expiration - date_evaluation).days
    date_evaluation = ql.Date(date_evaluation.day, date_evaluation.month,
                              date_evaluation.year)
    date_expiration = ql.Date(date_expiration.day, date_expiration.month,
                              date_expiration.year)

    # set the global evaluation date
    ql.Settings.instance().evaluationDate = date_evaluation

    return date_expiration, date_evaluation, dte


class Model():
    """Create a class used to model option prices

    For dates, must pass either 'dte' that is 'days to expiration',
    where the current date will be used as eval date and a number of
    days will be added to calculate an expiration date; alternatively
    can pass an expiration date and optionally an evaluation date (else
    current date is used).

    """
    def __init__(self, price_underlying, price_strike, volatility,
                 rate_risk_free, date_expiration=None, date_evaluation=None,
                 dte=None, option_type='call', verbose=False, dividend_rate=0,
                 exercise_type='european', n_steps=200):
        self.price_underlying = ql.SimpleQuote(price_underlying)
        self.price_strike = price_strike
        volatility, volatility_curve = build_volatility_curve(volatility)
        self.volatility = volatility
        self.volatility_curve = volatility_curve
        rate_risk_free_curve = build_interest_curve(rate_risk_free)
        self.n_steps = n_steps
        self.rate_risk_free = rate_risk_free
        self.rate_risk_free_curve = rate_risk_free_curve
        date_expiration, date_evaluation, dte = \
            build_dates(date_expiration, date_evaluation, dte)
        self.date_expiration = date_expiration
        self.date_evaluation = date_evaluation
        self.dte = dte
        self.verbose = verbose
        self.dividend_rate = dividend_rate
        self.day_count = ql.Actual365Fixed()
        self.calendar = ql.UnitedStates()
        self.exercise_type = exercise_type

        if option_type == 'call':
            self.option_type = ql.Option.Call
        elif option_type == 'C' or option_type == 'c':
            self.option_type = ql.Option.Call
        elif option_type == 'P' or option_type == 'p':
            self.option_type = ql.Option.Put
        else:
            self.option_type = ql.Option.Put
        self.create_process()
        self.create_option()

    def calculate_imp_vol(self, price_option, minvol=0.001, maxvol=10):
        """Use model to calculate volatility implied by price"""
        volatility_implied = \
            self.option.impliedVolatility(price_option, self.process,
                                          minVol=minvol, maxVol=10,)

        return volatility_implied

    def create_process(self):
        """Create the process"""
        # create process, engine and option
        self.process = ql.BlackScholesProcess(
            ql.QuoteHandle(self.price_underlying),
            ql.YieldTermStructureHandle(self.rate_risk_free_curve),
            ql.BlackVolTermStructureHandle(self.volatility_curve))

    def create_engine(self):
        if self.exercise_type == 'european':
            self.exercise = ql.EuropeanExercise(self.date_expiration)
            self.engine = ql.AnalyticEuropeanEngine(self.process)
        elif self.exercise_type == 'american':
            self.exercise = ql.AmericanExercise(self.date_evaluation,
                                                self.date_expiration)
            self.engine = ql.BinomialVanillaEngine(self.process, 'crr',
                                                   self.n_steps)
        else:
            raise Exception("Received unexpected exercise type "
                            f"'{self.exercise_type}'.")

        self.option = ql.VanillaOption(self.payoff, self.exercise)
        self.option.setPricingEngine(self.engine)

    def create_option(self):
        """Create the option"""
        self.create_process()
        self.payoff = ql.PlainVanillaPayoff(self.option_type, self.price_strike)
        self.create_engine()



import asyncclick as click


@click.group(chain=True, help='Buy, sell, and analyze futures.')
async def future():
    pass


@future.command()
@click.argument('quantity')
@click.argument('symbol')
@click.argument('price')
async def buy(quantity, symbol, price):
    """Buy QUANTITY contracts of SYMBOL.
    PRICE is the limit price per contract.
    """
    print(f'Buying {quantity}x {symbol} @ ${price}')


@future.command()
@click.argument('quantity')
@click.argument('symbol')
@click.argument('price')
async def sell(quantity, symbol, price):
    """Sell QUANTITY contracts of SYMBOL.
    PRICE is the limit price per contract.
    """
    print(f'Selling {quantity}x {symbol} @ ${price}')


@future.command()
@click.argument('symbol')
async def spot(symbol):
    """Look up the current bid/ask for SYMBOL.
    """
    print('Current price: Who the heck knows?')



import getpass
import logging
import os
from datetime import date, timedelta
from decimal import Decimal
from typing import Optional

import requests
from dateutil.relativedelta import FR, relativedelta
from tastyworks.models.session import TastyAPISession
from tastyworks.models.trading_account import TradingAccount

VERSION = '1.0.0'
ZERO = Decimal(0)
LOGGER = logging.getLogger(__name__)

_TOKEN_PATH = '.tastyworks/twcli/sesh'


class TastyworksCLIError(Exception):
    pass


class RenewableTastyAPISession(TastyAPISession):
    def __init__(self, API_url=None):
        path = os.path.join(os.path.expanduser('~'), _TOKEN_PATH)
        self.logged_in = False
        self.accounts = None

        # try to load token
        if os.path.exists(path):
            with open(path) as f:
                self.session_token = f.read().strip()

            self.API_url = API_url if API_url else 'https://api.tastyworks.com'

            # make sure token hasn't expired
            response = requests.post(f'{self.API_url}/sessions/validate', headers=self.get_request_headers())
            self.logged_in = (response.status_code == 201)

        if not self.logged_in:
            # either the token expired or doesn't exist
            username, password = self._get_credentials()
            TastyAPISession.__init__(self, username, password)

            # write session token to cache
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, 'w') as f:
                f.write(self.session_token)
        else:
            LOGGER.debug('Logged in with cached session token.')

    def _get_credentials(self):
        username = os.getenv('TW_USER')
        if not username:
            username = getpass.getpass('Username: ')
        password = os.getenv('TW_PASS')
        if not password:
            password = getpass.getpass('Password: ')

        return username, password

    @classmethod
    async def create(cls):
        self = RenewableTastyAPISession()
        accounts = await TradingAccount.get_remote_accounts(self)
        self.accounts = [acc for acc in accounts if not acc.is_closed]

        return self


def get_tasty_monthly(day: Optional[date] = date.today()) -> date:
    option1 = get_monthly(day + timedelta(weeks=4))
    option2 = get_monthly(day + timedelta(weeks=8))
    day45 = day + timedelta(days=45)
    return option1 if day45 - option1 < option2 - day45 else option2


def get_monthly(day: Optional[date] = date.today()) -> date:
    day = day.replace(day=1)
    day += relativedelta(weeks=2, weekday=FR)
    return day


def is_monthly(day: date) -> bool:
    return day.weekday() == 4 and 15 <= day.day <= 21


def get_confirmation(prompt: str) -> bool:
    while True:
        answer = input(prompt).lower()
        if not answer:
            return True
        if answer[0] == 'y':
            return True
        if answer[0] == 'n':
            return False


async def get_account(sesh: RenewableTastyAPISession) -> TradingAccount:
    accounts = await TradingAccount.get_remote_accounts(sesh)
    accounts = [acc for acc in accounts if not acc.is_closed]
    account = os.getenv('TW_ACC')
    if account:
        for acc in accounts:
            if acc.account_number == account:
                return acc
        LOGGER.warning('Environment variable $TW_ACC is set, but doesn\'t appear to exist!')

    for i in range(len(accounts)):
        if i == 0:
            print(f'{i + 1}) {accounts[i].account_number} {accounts[i].nickname} (default)')
        else:
            print(f'{i + 1}) {accounts[i].account_number} {accounts[i].nickname}')
    choice = 0
    while choice not in range(1, len(accounts) + 1):
        try:
            raw = input('Please choose an account: ')
            choice = int(raw)
        except ValueError:
            if not raw:
                return accounts[0]

    return accounts[choice - 1]



from setuptools import find_packages, setup

from twcli.utils import VERSION


f = open('README.md', 'r')
LONG_DESCRIPTION = f.read()
f.close()

setup(
    name='tastyworks-cli',
    version=VERSION,
    description='An easy-to-use command line interface for Tastyworks!',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Graeme Holliday',
    author_email='graeme.holliday@pm.me',
    url='https://github.com/Graeme22/tastyworks-cli/',
    license='MIT',
    install_requires=[
        'asyncclick>=8.0.3.2',
        'matplotlib>=3.5.1',
        'pandas>=1.4.0',
        'petl>=1.7.7',
        'python-dateutil>=2.8.2',
        'rich>=11.1.0',
        'QuantLib>=1.25',
        'tastyworks-api>=4.2.2',
    ],
    packages=find_packages(exclude=['ez_setup', 'tests*']),
    include_package_data=True,
    entry_points="""
        [console_scripts]
        twcli = twcli.app:main
    """,
)



import asyncclick as click

from .sizing import size as size_pair


def summarize(pair_specs):
    """Summarize pairs calculation"""
    print(f"\nFor {pair_specs['unit_size_right']:,.2f} units in the left "
          f"leg, trade {pair_specs['unit_size_left']:,.2f} in the right "
          f"leg. \n")


@click.group(chain=True, help='Evaluate and size pairs trades.')
async def pairs():
    pass


@pairs.command(help='Calculate volatility-weighted size for each leg. ')
@click.option('-i', '--interactive-mode', is_flag=True,
              help='Query inputs from user.')
@click.option('-l', '--price-left', type=float, default=1,
              help='Price of left leg.')
@click.option('-r', '--price-right', type=float, default=1,
              help='Price of right leg.')
@click.option('-v', '--vol-left', type=float, default=1,
              help='Volatility measure of left leg.')
@click.option('-o', '--vol-right', type=float, default=1,
              help='Volatility measure of right leg.')
@click.option('-s', '--unit-size-left', type=float, default=1,
              help='Number of units for left leg.')
@click.option('-m', '--multiplier-left', type=float, default=1,
              help='Unit multplier for left leg.')
@click.option('-u', '--multiplier-right', type=float, default=1,
              help='Unit multplier for right leg.')
async def size(interactive_mode, price_left, price_right, vol_left, vol_right,
               unit_size_left, multiplier_left, multiplier_right):
    if interactive_mode:
        args = dict()
        args['unit_size_left'] = 1
        args['multiplier_left'] = 1
        args['multiplier_right'] = 1
        c = ['unit_size_left', 'price_left', 'vol_left', 'multiplier_left',
             'price_right', 'vol_right', 'multiplier_right', ]
        for k in c:
            if k in args:
                v = input(f"enter value for {k} (default={args[k]:.2f}) > ")
                if v:
                    args[k] = float(v)
            else:
                args[k] = float(input(f"enter value for {k} > "))
        pair_specs = size_pair(**args)
        summarize(pair_specs)
        print('Below are the arguments to replicate: \n')
        print(
            f"    tw pairs size "
            f"-s {pair_specs['unit_size_left']:.2f} "
            f"-l {pair_specs['price_left']:.2f} "
            f"-v {pair_specs['vol_left']:.2f} "
            f"-m {pair_specs['multiplier_left']:.2f} "
            f"-r {pair_specs['price_right']:.2f} "
            f"-o {pair_specs['vol_right']:.2f} "
            f"-u {pair_specs['multiplier_right']:.2f} \n"
        )

    else:
        pair_specs = size_pair(price_left, price_right, vol_left, vol_right,
                               unit_size_left, multiplier_left,
                               multiplier_right)
        summarize(pair_specs)
